"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class OMEMOMessage(google.protobuf.message.Message):
    """A modified version of SignalMessage defined in
    https://github.com/signalapp/libsignal-protocol-java/blob/master/protobuf/WhisperTextProtocol.proto
    Structure, field names, required/optional and order changed to match python-twomemo. The serialization
    should not be affected by those modifications.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    N_FIELD_NUMBER: builtins.int
    PN_FIELD_NUMBER: builtins.int
    DH_PUB_FIELD_NUMBER: builtins.int
    CIPHERTEXT_FIELD_NUMBER: builtins.int
    n: builtins.int
    pn: builtins.int
    dh_pub: builtins.bytes
    ciphertext: builtins.bytes
    def __init__(self,
        *,
        n: typing.Optional[builtins.int] = ...,
        pn: typing.Optional[builtins.int] = ...,
        dh_pub: typing.Optional[builtins.bytes] = ...,
        ciphertext: typing.Optional[builtins.bytes] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["ciphertext",b"ciphertext","dh_pub",b"dh_pub","n",b"n","pn",b"pn"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["ciphertext",b"ciphertext","dh_pub",b"dh_pub","n",b"n","pn",b"pn"]) -> None: ...
global___OMEMOMessage = OMEMOMessage

class OMEMOKeyExchange(google.protobuf.message.Message):
    """A modified version of PreKeySignalMessage defined in
    https://github.com/signalapp/libsignal-protocol-java/blob/master/protobuf/WhisperTextProtocol.proto
    Structure, field names, required/optional and order changed to match python-twomemo. The serialization
    should not be affected by those modifications.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    PK_ID_FIELD_NUMBER: builtins.int
    SPK_ID_FIELD_NUMBER: builtins.int
    IK_FIELD_NUMBER: builtins.int
    EK_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    UNUSED_FIELD_NUMBER: builtins.int
    pk_id: builtins.int
    spk_id: builtins.int
    ik: builtins.bytes
    ek: builtins.bytes
    message: builtins.bytes
    """Byte-encoding of an OMEMOMessage & authentication tag (see OMEMOAuthenticatedMessage in oldmemo/oldmemo.py)"""

    unused: builtins.int
    def __init__(self,
        *,
        pk_id: typing.Optional[builtins.int] = ...,
        spk_id: typing.Optional[builtins.int] = ...,
        ik: typing.Optional[builtins.bytes] = ...,
        ek: typing.Optional[builtins.bytes] = ...,
        message: typing.Optional[builtins.bytes] = ...,
        unused: typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["ek",b"ek","ik",b"ik","message",b"message","pk_id",b"pk_id","spk_id",b"spk_id","unused",b"unused"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["ek",b"ek","ik",b"ik","message",b"message","pk_id",b"pk_id","spk_id",b"spk_id","unused",b"unused"]) -> None: ...
global___OMEMOKeyExchange = OMEMOKeyExchange



from __future__ import annotations

import base64
import secrets
from typing import Dict, NamedTuple, Optional, Tuple, cast
from typing_extensions import Final

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

import doubleratchet
from doubleratchet.recommended import (
    aead_aes_hmac,
    diffie_hellman_ratchet_curve25519,
    HashFunction,
    kdf_hkdf,
    kdf_separate_hmacs
)
from doubleratchet.recommended.crypto_provider_impl import CryptoProviderImpl
import google.protobuf.message
import xeddsa
import x3dh
import x3dh.identity_key_pair

from omemo.backend import Backend, DecryptionFailed, KeyExchangeFailed
from omemo.bundle import Bundle
from omemo.identity_key_pair import IdentityKeyPair, IdentityKeyPairSeed
from omemo.message import Content, EncryptedKeyMaterial, PlainKeyMaterial, KeyExchange
from omemo.session import Initiation, Session
from omemo.storage import Storage
from omemo.types import JSONType

# https://github.com/PyCQA/pylint/issues/4987
from .oldmemo_pb2 import (  # pylint: disable=no-name-in-module
    OMEMOKeyExchange,
    OMEMOMessage
)


__all__ = [
    "Oldmemo",
    "NAMESPACE",
    "AEADImpl",
    "BundleImpl",
    "ContentImpl",
    "DoubleRatchetImpl",
    "EncryptedKeyMaterialImpl",
    "KeyExchangeImpl",
    "MessageChainKDFImpl",
    "PlainKeyMaterialImpl",
    "RootChainKDFImpl",
    "SessionImpl",
    "StateImpl"
]


NAMESPACE: Final = "eu.siacs.conversations.axolotl"


class RootChainKDFImpl(kdf_hkdf.KDF):
    """
    The root chain KDF implementation used by this version of the specification.
    """

    @staticmethod
    def _get_hash_function() -> HashFunction:
        return HashFunction.SHA_256

    @staticmethod
    def _get_info() -> bytes:
        # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/
        # java/src/main/java/org/whispersystems/libsignal/ratchet/RootKey.java#L35
        return "WhisperRatchet".encode("ASCII")


class MessageChainKDFImpl(kdf_separate_hmacs.KDF):
    """
    The message chain KDF implementation used by this version of the specification.
    """

    @staticmethod
    def _get_hash_function() -> HashFunction:
        return HashFunction.SHA_256


class OMEMOAuthenticatedMessage(NamedTuple):
    # pylint: disable=invalid-name
    """
    The `urn:xmpp:omemo:2` version of the specification uses a protobuf structure called
    ``OMEMOAuthenticatedMessage`` to hold and transfer a serialized :class:`OMEMOMessage` in network format
    and an authentication tag. This version of the specification instead uses simple concatenation of the two
    byte strings. This class mocks the protobuf structure API to make maintaining this backend as a fork of
    python-twomemo easier and to make the code cleaner.
    """

    mac: bytes
    message: bytes

    def SerializeToString(self, deterministic: bool = True) -> bytes:
        """
        Args:
            deterministic: This parameter only exists to mimic protobuf's structure API and must be ``True``.

        Returns:
            The contents of this instance serialized as a byte string.
        """

        assert deterministic

        return self.message + self.mac

    @staticmethod
    def FromString(serialized: bytes) -> OMEMOAuthenticatedMessage:
        """
        Args:
            serialized: A serialized instance as returned by :meth:`SerializeToString`.

        Returns:
            An instance with the data restored from the serialized input.
        """

        return OMEMOAuthenticatedMessage(
            mac=serialized[-AEADImpl.AUTHENTICATION_TAG_TRUNCATED_LENGTH:],
            message=serialized[:-AEADImpl.AUTHENTICATION_TAG_TRUNCATED_LENGTH]
        )


class AEADImpl(aead_aes_hmac.AEAD):
    """
    The AEAD used by this backend as part of the Double Ratchet. While this implementation derives from
    :class:`doubleratchet.recommended.aead_aes_hmac.AEAD`, it actually doesn't use any of its code. This is
    due to a minor difference in the way the associated data is built. The derivation only has symbolic value.

    Can only be used with :class:`DoubleRatchetImpl`, due to the reliance on a certain structure of the
    associated data.
    """

    # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/java/
    # src/main/java/org/whispersystems/libsignal/protocol/SignalMessage.java#L28
    AUTHENTICATION_TAG_TRUNCATED_LENGTH: Final = 8

    @staticmethod
    def _get_hash_function() -> HashFunction:
        return HashFunction.SHA_256

    @staticmethod
    def _get_info() -> bytes:
        # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/
        # java/src/main/java/org/whispersystems/libsignal/ratchet/ChainKey.java#L48
        return "WhisperMessageKeys".encode("ASCII")

    @classmethod
    async def encrypt(cls, plaintext: bytes, key: bytes, associated_data: bytes) -> bytes:
        hash_function = cls._get_hash_function()

        encryption_key, authentication_key, iv = await cls.__derive(key, hash_function, cls._get_info())

        # Encrypt the plaintext using AES-256 (the 256 bit are implied by the key size) in CBC mode and the
        # previously created key and IV, after padding it with PKCS#7
        ciphertext = await CryptoProviderImpl.aes_cbc_encrypt(encryption_key, iv, plaintext)

        # Parse the associated data
        associated_data, header = cls.__parse_associated_data(associated_data)

        # Build an OMEMOMessage including the header and the ciphertext
        omemo_message = b"\x33" + OMEMOMessage(
            n=header.sending_chain_length,
            pn=header.previous_sending_chain_length,
            dh_pub=StateImpl.serialize_public_key(header.ratchet_pub),
            ciphertext=ciphertext
        ).SerializeToString()

        # Calculate the authentication tag over the associated data and the OMEMOMessage, truncate the
        # authentication tag to AUTHENTICATION_TAG_TRUNCATED_LENGTH bytes
        auth = (await CryptoProviderImpl.hmac_calculate(
            authentication_key,
            hash_function,
            associated_data + omemo_message
        ))[:AEADImpl.AUTHENTICATION_TAG_TRUNCATED_LENGTH]

        # Serialize the authentication tag with the OMEMOMessage in an OMEMOAuthenticatedMessage.
        return OMEMOAuthenticatedMessage(mac=auth, message=omemo_message).SerializeToString()

    @classmethod
    async def decrypt(cls, ciphertext: bytes, key: bytes, associated_data: bytes) -> bytes:
        hash_function = cls._get_hash_function()

        decryption_key, authentication_key, iv = await cls.__derive(key, hash_function, cls._get_info())

        # Parse the associated data
        associated_data, header = cls.__parse_associated_data(associated_data)

        # Parse the ciphertext as an OMEMOAuthenticatedMessage
        try:
            omemo_authenticated_message = OMEMOAuthenticatedMessage.FromString(ciphertext)
        except google.protobuf.message.DecodeError as e:
            raise doubleratchet.DecryptionFailedException() from e

        # Calculate and verify the authentication tag
        new_auth = (await CryptoProviderImpl.hmac_calculate(
            authentication_key,
            hash_function,
            associated_data + omemo_authenticated_message.message
        ))[:AEADImpl.AUTHENTICATION_TAG_TRUNCATED_LENGTH]

        if new_auth != omemo_authenticated_message.mac:
            raise doubleratchet.aead.AuthenticationFailedException("Authentication tags do not match.")

        # Parse the OMEMOMessage contained in the OMEMOAuthenticatedMessage
        if len(omemo_authenticated_message.message) < 1 or omemo_authenticated_message.message[0] != 0x33:
            raise doubleratchet.DecryptionFailedException("Version byte missing.")

        try:
            omemo_message = OMEMOMessage.FromString(omemo_authenticated_message.message[1:])
        except google.protobuf.message.DecodeError as e:
            raise doubleratchet.DecryptionFailedException() from e

        # Make sure that the headers match as a little additional consistency check
        if header != doubleratchet.Header(
            StateImpl.parse_public_key(omemo_message.dh_pub),
            omemo_message.pn,
            omemo_message.n
        ):
            raise doubleratchet.aead.AuthenticationFailedException("Header mismatch.")

        # Decrypt the plaintext using AES-256 (the 256 bit are implied by the key size) in CBC mode and the
        # previously created key and IV, and unpad the resulting plaintext with PKCS#7
        return await CryptoProviderImpl.aes_cbc_decrypt(decryption_key, iv, omemo_message.ciphertext)

    @staticmethod
    async def __derive(key: bytes, hash_function: HashFunction, info: bytes) -> Tuple[bytes, bytes, bytes]:
        # Prepare the salt, a zero-filled byte sequence with the size of the hash digest
        salt = b"\x00" * hash_function.hash_size

        # Derive 80 bytes
        hkdf_out = await CryptoProviderImpl.hkdf_derive(
            hash_function=hash_function,
            length=80,
            salt=salt,
            info=info,
            key_material=key
        )

        # Split these 80 bytes into three parts
        return hkdf_out[:32], hkdf_out[32:64], hkdf_out[64:]

    @staticmethod
    def __parse_associated_data(associated_data: bytes) -> Tuple[bytes, doubleratchet.Header]:
        """
        Parse the associated data as built by :meth:`DoubleRatchetImpl._build_associated_data`.

        Args:
            associated_data: The associated data.

        Returns:
            The original associated data and the header used to build it.

        Raises:
            doubleratchet.DecryptionFailedException: if the data is malformed.
        """

        associated_data_length = StateImpl.IDENTITY_KEY_ENCODING_LENGTH * 2

        try:
            omemo_message = OMEMOMessage.FromString(associated_data[associated_data_length:])
        except google.protobuf.message.DecodeError as e:
            raise doubleratchet.DecryptionFailedException() from e

        associated_data = associated_data[:associated_data_length]

        return associated_data, doubleratchet.Header(
            StateImpl.parse_public_key(omemo_message.dh_pub),
            omemo_message.pn,
            omemo_message.n
        )


class DoubleRatchetImpl(doubleratchet.DoubleRatchet):
    """
    The Double Ratchet implementation used by this version of the specification.
    """

    # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/java/
    # src/main/java/org/whispersystems/libsignal/ratchet/ChainKey.java#L20
    MESSAGE_CHAIN_CONSTANT: Final = b"\x02\x01"

    @staticmethod
    def _build_associated_data(associated_data: bytes, header: doubleratchet.Header) -> bytes:
        return associated_data + OMEMOMessage(
            n=header.sending_chain_length,
            pn=header.previous_sending_chain_length,
            dh_pub=StateImpl.serialize_public_key(header.ratchet_pub)
        ).SerializeToString()


class StateImpl(x3dh.BaseState):
    """
    The X3DH state implementation used by this version of the specification.
    """

    # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/java/
    # src/main/java/org/whispersystems/libsignal/ratchet/RatchetingSession.java#L132
    INFO: Final = "WhisperText".encode("ASCII")

    IDENTITY_KEY_ENCODING_LENGTH: Final = 33  # One byte constant + 32 bytes key

    @staticmethod
    def _encode_public_key(key_format: x3dh.IdentityKeyFormat, pub: bytes) -> bytes:
        # python-omemo uses Ed25519 for the identity key format, while the 0.3.0 specification uses Curve25519
        # encoding. This is one of the places where this difference matters, since the identity key will be
        # passed in Ed25519 format, but required to be encoded as Curve25519. All keys but the identity key
        # are fixed to Curve25519 format anyway, thus the following check should handle the compatibility:
        if key_format is x3dh.IdentityKeyFormat.ED_25519:
            pub = xeddsa.ed25519_pub_to_curve25519_pub(pub)

        # https://github.com/signalapp/libsignal-protocol-java/blob/fde96d22004f32a391554e4991e4e1f0a14c2d50/
        # java/src/main/java/org/whispersystems/libsignal/ecc/Curve.java#L17
        return b"\x05" + pub

    @staticmethod
    def serialize_public_key(pub: bytes) -> bytes:
        """
        Args:
            pub: A public key in Curve25519 format.

        Returns:
            The public key serialized in the network format.

        Note:
            This is a reexport of :meth:`_encode_public_key` with the format fixed to Curve25519.
        """

        return StateImpl._encode_public_key(x3dh.IdentityKeyFormat.CURVE_25519, pub)

    @staticmethod
    def parse_public_key(serialized: bytes) -> bytes:
        """
        Args:
            serialized: A Curve25519 public key serialized in the network format, as returned by e.g.
                :meth:`serialize_public_key`.

        Returns:
            The parsed public key in Curve25519 format.

        Raises:
            ValueError: if the input format does not comply to the expected network format.
        """

        if len(serialized) == StateImpl.IDENTITY_KEY_ENCODING_LENGTH and serialized[0] == 0x05:
            return serialized[1:]

        raise ValueError("Public key not serialized in network format.")


class BundleImpl(Bundle):
    """
    :class:`~omemo.bundle.Bundle` implementation as a simple storage type.
    """

    def __init__(
        self,
        bare_jid: str,
        device_id: int,
        bundle: x3dh.Bundle,
        signed_pre_key_id: int,
        pre_key_ids: Dict[bytes, int]
    ) -> None:
        """
        Args:
            bare_jid: The bare JID this bundle belongs to.
            device_id: The device id of the specific device this bundle belongs to.
            bundle: The bundle to store in this instance.
            signed_pre_key_id: The id of the signed pre key referenced in the bundle.
            pre_key_ids: A dictionary that maps each pre key referenced in the bundle to its id.
        """

        self.__bare_jid = bare_jid
        self.__device_id = device_id
        self.__bundle = bundle
        self.__signed_pre_key_id = signed_pre_key_id
        self.__pre_key_ids = dict(pre_key_ids)

    @property
    def namespace(self) -> str:
        return NAMESPACE

    @property
    def bare_jid(self) -> str:
        return self.__bare_jid

    @property
    def device_id(self) -> int:
        return self.__device_id

    @property
    def identity_key(self) -> bytes:
        return self.__bundle.identity_key

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BundleImpl):
            return (
                other.bare_jid == self.bare_jid
                and other.device_id == self.device_id
                and other.bundle == self.bundle
                and other.signed_pre_key_id == self.signed_pre_key_id
                and other.pre_key_ids == self.pre_key_ids
            )

        return False

    def __hash__(self) -> int:
        return hash((
            self.bare_jid,
            self.device_id,
            self.bundle,
            self.signed_pre_key_id,
            frozenset(self.pre_key_ids.items())
        ))

    @property
    def bundle(self) -> x3dh.Bundle:
        """
        Returns:
            The bundle held by this instance.
        """

        return self.__bundle

    @property
    def signed_pre_key_id(self) -> int:
        """
        Returns:
            The id of the signed pre key referenced in the bundle.
        """

        return self.__signed_pre_key_id

    @property
    def pre_key_ids(self) -> Dict[bytes, int]:
        """
        Returns:
            A dictionary that maps each pre key referenced in the bundle to its id.
        """

        return dict(self.__pre_key_ids)


class ContentImpl(Content):
    """
    :class:`~omemo.message.Content` implementation as a simple storage type.
    """

    def __init__(self, ciphertext: bytes, initialization_vector: bytes) -> None:
        """
        Args:
            ciphertext: The ciphertext to store in this instance.
            initialization_vector: The initialization vector to store in this instance.

        Note:
            For empty OMEMO messages as per the specification, the ciphertext is set to an empty byte string
            and the initialization vector is initialized with a valid initialization vector for further use by
            external protocols (aka ``KeyTransportMessage``).
        """

        self.__ciphertext = ciphertext
        self.__initialization_vector = initialization_vector

    @property
    def empty(self) -> bool:
        return self.__ciphertext == b""

    @staticmethod
    def make_empty() -> ContentImpl:
        """
        Returns:
            An "empty" instance, i.e. one that corresponds to an empty OMEMO message as per the specification.
            The ciphertext is set to an empty byte string and the initialization vector is initialized with a
            valid initialization vector for further use by external protocols (aka ``KeyTransportMessage``).
        """

        return ContentImpl(b"", secrets.token_bytes(12))

    @property
    def ciphertext(self) -> bytes:
        """
        Returns:
            The ciphertext held by this instance.
        """

        return self.__ciphertext

    @property
    def initialization_vector(self) -> bytes:
        """
        Returns:
            The initialization vector held by this instance.
        """

        return self.__initialization_vector


class EncryptedKeyMaterialImpl(EncryptedKeyMaterial):
    """
    :class:`~omemo.message.EncryptedKeyMaterial` implementation as a simple storage type.
    """

    def __init__(
        self,
        bare_jid: str,
        device_id: int,
        encrypted_message: doubleratchet.EncryptedMessage
    ) -> None:
        """
        Args:
            bare_jid: The bare JID of the other party.
            device_id: The device id of the specific device of the other party.
            encrypted_message: The encrypted Double Ratchet message to store in this instance.
        """

        self.__bare_jid = bare_jid
        self.__device_id = device_id
        self.__encrypted_message = encrypted_message

    @property
    def bare_jid(self) -> str:
        return self.__bare_jid

    @property
    def device_id(self) -> int:
        return self.__device_id

    @property
    def encrypted_message(self) -> doubleratchet.EncryptedMessage:
        """
        Returns:
            The encrypted Double Ratchet message held by this instance.
        """

        return self.__encrypted_message

    def serialize(self) -> bytes:
        """
        Returns:
            A serialized OMEMOAuthenticatedMessage message structure representing the content of this
            instance.
        """

        # The ciphertext field contains the result of :meth:`AEADImpl.encrypt`, which is a serialized
        # OMEMOAuthenticatedMessage with all fields already correctly set, thus it can be used here as is.
        return self.__encrypted_message.ciphertext

    @staticmethod
    def parse(authenticated_message: bytes, bare_jid: str, device_id: int) -> EncryptedKeyMaterialImpl:
        """
        Args:
            authenticated_message: A serialized OMEMOAuthenticatedMessage message structure.
            bare_jid: The bare JID of the other party.
            device_id: The device id of the specific device of the other party.

        Returns:
            An instance of this class, parsed from the OMEMOAuthenticatedMessage.

        Raises:
            ValueError: if the data is malformed.
        """

        message_serialized = OMEMOAuthenticatedMessage.FromString(authenticated_message).message

        if len(message_serialized) < 1 or message_serialized[0] != 0x33:
            raise ValueError("Version byte missing.")

        # Parse the OMEMOAuthenticatedMessage and OMEMOMessage structures to extract the header.
        try:
            message = OMEMOMessage.FromString(message_serialized[1:])
        except google.protobuf.message.DecodeError as e:
            raise ValueError() from e

        return EncryptedKeyMaterialImpl(
            bare_jid,
            device_id,
            doubleratchet.EncryptedMessage(
                doubleratchet.Header(
                    StateImpl.parse_public_key(message.dh_pub),
                    message.pn,
                    message.n
                ),
                authenticated_message
            )
        )


class PlainKeyMaterialImpl(PlainKeyMaterial):
    """
    :class:`~omemo.message.PlainKeyMaterial` implementation as a simple storage type.
    """

    KEY_LENGTH: Final = 16

    def __init__(self, key: bytes, auth_tag: bytes) -> None:
        """
        Args:
            key: The key to store in this instance.
            auth_tag: The authentication tag to store in this instance.

        Note:
            For empty OMEMO messages as per the specification, the key is set to a freshly generated key for
            further use by external protocols (aka ``KeyTransportMessage``), while the auth tag is set to an
            empty byte string.
        """

        self.__key = key
        self.__auth_tag = auth_tag

    @property
    def key(self) -> bytes:
        """
        Returns:
            The key held by this instance.
        """

        return self.__key

    @property
    def auth_tag(self) -> bytes:
        """
        Returns:
            The authentication tag held by this instance.
        """

        return self.__auth_tag

    @staticmethod
    def make_empty() -> PlainKeyMaterialImpl:
        """
        Returns:
            An "empty" instance, i.e. one that corresponds to an empty OMEMO message as per the specification.
            The key stored in empty instances is a freshly generated key for further use by external protocols
            (aka ``KeyTransportMessage``), while the auth tag is set to an empty byte string.
        """

        return PlainKeyMaterialImpl(secrets.token_bytes(PlainKeyMaterialImpl.KEY_LENGTH), b"")


class KeyExchangeImpl(KeyExchange):
    """
    :class:`~omemo.message.KeyExchange` implementation as a simple storage type.

    There are four kinds of instances:

    - Completely filled instances
    - Partially filled instances received via network
    - Very sparsely filled instances migrated from the legacy storage format
    - Almost completely empty instances migrated from the legacy storage format

    Empty fields are filled with filler values such that the data types and lengths still match expectations.
    The fourth kind, almost completely empty instances, will never have any of their methods called except for
    getters.
    """

    def __init__(self, header: x3dh.Header, signed_pre_key_id: int, pre_key_id: int) -> None:
        """
        Args:
            header: The header to store in this instance.
            signed_pre_key_id: The id of the signed pre key referenced in the header.
            pre_key_id: The id of the pre key referenced in the header.
        """

        self.__header = header
        self.__signed_pre_key_id = signed_pre_key_id
        self.__pre_key_id = pre_key_id

    @property
    def identity_key(self) -> bytes:
        return self.__header.identity_key

    def builds_same_session(self, other: KeyExchange) -> bool:
        if isinstance(other, KeyExchangeImpl):
            if self.is_migrated_instance() or other.is_migrated_instance():
                # If any of the instances is a migrated instance, we can only compare the identity key and the
                # pre key id. Sadly that's the only data included in the legacy storage format, next to the
                # pre key byte data, which doesn't add any value.
                return (
                    other.header.identity_key == self.header.identity_key
                    and other.pre_key_id == self.pre_key_id
                )

            # Otherwise, we are dealing with completely filled instances or network instances. The signed pre
            # key id and pre key id are enough for uniqueness; ignoring the actual signed pre key and pre key
            # bytes here makes it possible to compare network instances with completely filled instances.
            return (
                other.header.identity_key == self.header.identity_key
                and other.header.ephemeral_key == self.header.ephemeral_key
                and other.signed_pre_key_id == self.signed_pre_key_id
                and other.pre_key_id == self.pre_key_id
            )

        return False

    @property
    def header(self) -> x3dh.Header:
        """
        Returns:
            The header held by this instance.
        """

        return self.__header

    @property
    def signed_pre_key_id(self) -> int:
        """
        Returns:
            The id of the signed pre key referenced in the header.
        """

        return self.__signed_pre_key_id

    @property
    def pre_key_id(self) -> int:
        """
        Returns:
            The id of the pre key referenced in the header.
        """

        return self.__pre_key_id

    def is_network_instance(self) -> bool:
        """
        Returns:
            Returns whether this is a network instance. A network instance has all fields filled except for
            the signed pre key and pre key byte data. The missing byte data can be restored by looking it up
            from storage using the respective ids.
        """

        return self.__header.signed_pre_key == b"" and self.__header.pre_key == b""

    def is_migrated_instance(self) -> bool:
        """
        Returns:
            Whether this is a migrated instance, according to the third kind as described in the class
            docstring. A migrated instance of that kind only sets the identity key, the pre key id and the pre
            key byte data. Other values are fillers.
        """

        # Could confirm the other values here too, but why the trouble.
        return self.__signed_pre_key_id == -1 and self.__pre_key_id != -1

    def serialize(self, authenticated_message: bytes) -> Tuple[bytes, bool]:
        """
        Args:
            authenticated_message: The serialized OMEMOAuthenticatedMessage message structure to include with
                the key exchange information.

        Returns:
            A serialized OMEMOKeyExchange message structure in network format representing the content of this
            instance, and a flag indicating whether the sign bit was is on the identity key in its Ed25519
            form.
        """

        return b"\x33" + OMEMOKeyExchange(
            pk_id=self.__pre_key_id,
            spk_id=self.__signed_pre_key_id,
            ik=StateImpl.serialize_public_key(xeddsa.ed25519_pub_to_curve25519_pub(
                self.__header.identity_key
            )),
            ek=StateImpl.serialize_public_key(self.__header.ephemeral_key),
            message=authenticated_message
        ).SerializeToString(), bool((self.__header.identity_key[31] >> 7) & 1)

    @staticmethod
    def parse(key_exchange: bytes, set_sign_bit: bool) -> Tuple[KeyExchangeImpl, bytes]:
        """
        Args:
            key_exchange: A serialized OMEMOKeyExchange message structure in network format.
            set_sign_bit: Whether to set the sign bit on the identity key when converting it to its Ed25519
                form.

        Returns:
            An instance of this class, parsed from the OMEMOKeyExchange, and the serialized
            OMEMOAuthenticatedMessage extracted from the OMEMOKeyExchange.

        Raises:
            ValueError: if the data is malformed.

        Warning:
            The OMEMOKeyExchange message structure only contains the ids of the signed pre key and the pre key
            used for the key exchange, not the full public keys. Since the job of this method is just parsing,
            the X3DH header is initialized without the public keys here, and the code using instances of this
            class has to handle the public key lookup from the ids. Use :meth:`is_network_instance` to check
            whether the header is a full or a partial (network) instance.
        """

        if len(key_exchange) < 1 or key_exchange[0] != 0x33:
            raise ValueError("Version byte missing.")

        key_exchange = key_exchange[1:]

        try:
            parsed = OMEMOKeyExchange.FromString(key_exchange)
        except google.protobuf.message.DecodeError as e:
            raise ValueError() from e

        return KeyExchangeImpl(
            x3dh.Header(
                xeddsa.curve25519_pub_to_ed25519_pub(StateImpl.parse_public_key(parsed.ik), set_sign_bit),
                StateImpl.parse_public_key(parsed.ek),
                b"",
                b""
            ),
            parsed.spk_id,
            parsed.pk_id
        ), parsed.message


class SessionImpl(Session):
    """
    :class:`~omemo.session.Session` implementation as a simple storage type.
    """

    def __init__(
        self,
        bare_jid: str,
        device_id: int,
        initiation: Initiation,
        key_exchange: KeyExchangeImpl,
        associated_data: bytes,
        double_ratchet: DoubleRatchetImpl,
        confirmed: bool = False
    ):
        """
        Args:
            bare_jid: The bare JID of the other party.
            device_id: The device id of the specific device of the other party.
            initiation: Whether this session was built through active or passive session initiation.
            key_exchange: The key exchange information to store in this instance.
            associated_data: The associated data to store in this instance.
            double_ratchet: The Double Ratchet to store in this instance.
            confirmed: Whether the session was confirmed, i.e. whether a message was decrypted after actively
                initiating the session. Leave this at the default value for passively initiated sessions.
        """

        self.__bare_jid = bare_jid
        self.__device_id = device_id
        self.__initiation = initiation
        self.__key_exchange = key_exchange
        self.__associated_data = associated_data
        self.__double_ratchet = double_ratchet
        self.__confirmed = confirmed

    @property
    def namespace(self) -> str:
        return NAMESPACE

    @property
    def bare_jid(self) -> str:
        return self.__bare_jid

    @property
    def device_id(self) -> int:
        return self.__device_id

    @property
    def initiation(self) -> Initiation:
        return self.__initiation

    @property
    def confirmed(self) -> bool:
        return self.__confirmed

    @property
    def key_exchange(self) -> KeyExchangeImpl:
        return self.__key_exchange

    @property
    def receiving_chain_length(self) -> Optional[int]:
        return self.__double_ratchet.receiving_chain_length

    @property
    def sending_chain_length(self) -> int:
        return self.__double_ratchet.sending_chain_length

    @property
    def associated_data(self) -> bytes:
        """
        Returns:
            The associated data held by this instance.
        """

        return self.__associated_data

    @property
    def double_ratchet(self) -> DoubleRatchetImpl:
        """
        Returns:
            The Double Ratchet held by this instance.
        """

        return self.__double_ratchet

    def confirm(self) -> None:
        """
        Mark this session as confirmed.
        """

        self.__confirmed = True


class Oldmemo(Backend):
    """
    :class:`~omemo.backend.Backend` implementation providing OMEMO in the `eu.siacs.conversations.axolotl`
    namespace.

    One notable implementation detail is the handling of the identity key format. The specification requires
    the identity key to be transferred in Curve25519 format (in bundles, key exchanges etc.), while the
    python-omemo library uses Ed25519 serialization whenever the identity key is referred to. Thus, conversion
    has to happen during the serialization/parsing of transferred data, as done for example in
    :mod:`oldmemo.etree`.
    """

    def __init__(
        self,
        storage: Storage,
        max_num_per_session_skipped_keys: int = 1000,
        max_num_per_message_skipped_keys: Optional[int] = None
    ) -> None:
        """
        Args:
            storage: The storage to store backend-specific data in. Note that all data keys are prefixed with
                the backend namespace to avoid name clashes between backends.
            max_num_per_session_skipped_keys: The maximum number of skipped message keys to keep around per
                session. Once the maximum is reached, old message keys are deleted to make space for newer
                ones. Accessible via :attr:`~omemo.backend.Backend.max_num_per_session_skipped_keys`.
            max_num_per_message_skipped_keys: The maximum number of skipped message keys to accept in a single
                message. When set to ``None`` (the default), this parameter defaults to the per-session
                maximum (i.e. the value of the ``max_num_per_session_skipped_keys`` parameter). This parameter
                may only be 0 if the per-session maximum is 0, otherwise it must be a number between 1 and the
                per-session maximum. Accessible via
                :attr:`~omemo.backend.Backend.max_num_per_message_skipped_keys`.
        """

        super().__init__(max_num_per_session_skipped_keys, max_num_per_message_skipped_keys)

        self.__storage = storage

    async def __get_state(self) -> StateImpl:
        """
        Returns:
            The loaded or newly created X3DH state.
        """

        def check_type(value: JSONType) -> x3dh.types.JSONObject:
            if isinstance(value, dict):
                return cast(x3dh.types.JSONObject, value)

            raise TypeError(
                f"Stored StateImpl under key /{self.namespace}/x3dh corrupt: not a JSON object: {value}"
            )

        state, _ = (await self.__storage.load(
            f"/{self.namespace}/x3dh"
        )).fmap(check_type).fmap(lambda serialized: StateImpl.from_json(
            serialized,
            x3dh.IdentityKeyFormat.ED_25519,
            x3dh.HashFunction.SHA_256,
            StateImpl.INFO
        )).maybe((None, False))

        if state is None:
            identity_key_pair = await IdentityKeyPair.get(self.__storage)

            state = StateImpl.create(
                x3dh.IdentityKeyFormat.ED_25519,
                x3dh.HashFunction.SHA_256,
                StateImpl.INFO,
                (
                    x3dh.identity_key_pair.IdentityKeyPairSeed(identity_key_pair.seed)
                    if isinstance(identity_key_pair, IdentityKeyPairSeed)
                    else x3dh.identity_key_pair.IdentityKeyPairPriv(identity_key_pair.as_priv().priv)
                )
            )

            await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

        return state

    @property
    def namespace(self) -> str:
        return NAMESPACE

    async def load_session(self, bare_jid: str, device_id: int) -> Optional[SessionImpl]:
        def check_type(value: JSONType) -> doubleratchet.types.JSONObject:
            if isinstance(value, dict):
                return cast(doubleratchet.types.JSONObject, value)

            raise TypeError(
                f"Stored DoubleRatchetImpl under key"
                f" /{self.namespace}/{bare_jid}/{device_id}/double_ratchet corrupt: not a JSON object:"
                f" {value}"
            )

        try:
            double_ratchet = (await self.__storage.load(
                f"/{self.namespace}/{bare_jid}/{device_id}/double_ratchet"
            )).fmap(check_type).fmap(lambda serialized: DoubleRatchetImpl.from_json(
                serialized,
                diffie_hellman_ratchet_curve25519.DiffieHellmanRatchet,
                RootChainKDFImpl,
                MessageChainKDFImpl,
                DoubleRatchetImpl.MESSAGE_CHAIN_CONSTANT,
                self.max_num_per_message_skipped_keys,
                self.max_num_per_session_skipped_keys,
                AEADImpl
            )).maybe(None)
        except doubleratchet.InconsistentSerializationException:
            return None

        if double_ratchet is None:
            return None

        initiation = Initiation((await self.__storage.load_primitive(
            f"/{self.namespace}/{bare_jid}/{device_id}/initiation",
            str
        )).from_just())

        identity_key = (await self.__storage.load_bytes(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/identity_key"
        )).from_just()

        ephemeral_key = (await self.__storage.load_bytes(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/ephemeral_key"
        )).from_just()

        signed_pre_key = (await self.__storage.load_bytes(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/signed_pre_key"
        )).from_just()

        signed_pre_key_id = (await self.__storage.load_primitive(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/signed_pre_key_id",
            int
        )).from_just()

        pre_key = (await self.__storage.load_bytes(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/pre_key"
        )).from_just()

        pre_key_id = (await self.__storage.load_primitive(
            f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/pre_key_id",
            int
        )).from_just()

        associated_data = (await self.__storage.load_bytes(
            f"/{self.namespace}/{bare_jid}/{device_id}/associated_data"
        )).from_just()

        confirmed = (await self.__storage.load_primitive(
            f"/{self.namespace}/{bare_jid}/{device_id}/confirmed",
            bool
        )).from_just()

        return SessionImpl(bare_jid, device_id, initiation, KeyExchangeImpl(
            x3dh.Header(identity_key, ephemeral_key, signed_pre_key, pre_key),
            signed_pre_key_id,
            pre_key_id
        ), associated_data, double_ratchet, confirmed)

    async def store_session(self, session: Session) -> None:
        assert isinstance(session, SessionImpl)

        assert session.key_exchange.header.pre_key is not None

        await self.__storage.store(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/initiation",
            session.initiation.name
        )

        await self.__storage.store_bytes(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/identity_key",
            session.key_exchange.header.identity_key
        )

        await self.__storage.store_bytes(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/ephemeral_key",
            session.key_exchange.header.ephemeral_key
        )

        await self.__storage.store_bytes(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/signed_pre_key",
            session.key_exchange.header.signed_pre_key
        )

        await self.__storage.store(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/signed_pre_key_id",
            session.key_exchange.signed_pre_key_id
        )

        await self.__storage.store_bytes(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/pre_key",
            session.key_exchange.header.pre_key
        )

        await self.__storage.store(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/pre_key_id",
            session.key_exchange.pre_key_id
        )

        await self.__storage.store_bytes(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/associated_data",
            session.associated_data
        )

        await self.__storage.store(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/double_ratchet",
            session.double_ratchet.json
        )

        await self.__storage.store(
            f"/{self.namespace}/{session.bare_jid}/{session.device_id}/confirmed",
            session.confirmed
        )

        # Keep track of bare JIDs with stored sessions
        bare_jids = set((await self.__storage.load_list(f"/{self.namespace}/bare_jids", str)).maybe([]))
        bare_jids.add(session.bare_jid)
        await self.__storage.store(f"/{self.namespace}/bare_jids", list(bare_jids))

        # Keep track of device ids with stored sessions
        device_ids = set((await self.__storage.load_list(
            f"/{self.namespace}/{session.bare_jid}/device_ids",
            int
        )).maybe([]))
        device_ids.add(session.device_id)
        await self.__storage.store(f"/{self.namespace}/{session.bare_jid}/device_ids", list(device_ids))

    async def build_session_active(
        self,
        bare_jid: str,
        device_id: int,
        bundle: Bundle,
        plain_key_material: PlainKeyMaterial
    ) -> Tuple[SessionImpl, EncryptedKeyMaterialImpl]:
        assert isinstance(bundle, BundleImpl)
        assert isinstance(plain_key_material, PlainKeyMaterialImpl)

        try:
            state = await self.__get_state()
            shared_secret, associated_data, header = await state.get_shared_secret_active(bundle.bundle)
        except x3dh.KeyAgreementException as e:
            raise KeyExchangeFailed() from e

        assert header.pre_key is not None

        double_ratchet, encrypted_message = await DoubleRatchetImpl.encrypt_initial_message(
            diffie_hellman_ratchet_curve25519.DiffieHellmanRatchet,
            RootChainKDFImpl,
            MessageChainKDFImpl,
            DoubleRatchetImpl.MESSAGE_CHAIN_CONSTANT,
            self.max_num_per_message_skipped_keys,
            self.max_num_per_session_skipped_keys,
            AEADImpl,
            shared_secret,
            bundle.bundle.signed_pre_key,
            plain_key_material.key + plain_key_material.auth_tag,
            associated_data
        )

        session = SessionImpl(
            bare_jid,
            device_id,
            Initiation.ACTIVE,
            KeyExchangeImpl(
                header,
                bundle.signed_pre_key_id,
                bundle.pre_key_ids[header.pre_key]
            ),
            associated_data,
            double_ratchet
        )

        encrypted_key_material = EncryptedKeyMaterialImpl(bare_jid, device_id, encrypted_message)

        return session, encrypted_key_material

    async def build_session_passive(
        self,
        bare_jid: str,
        device_id: int,
        key_exchange: KeyExchange,
        encrypted_key_material: EncryptedKeyMaterial
    ) -> Tuple[SessionImpl, PlainKeyMaterialImpl]:
        assert isinstance(key_exchange, KeyExchangeImpl)
        assert isinstance(encrypted_key_material, EncryptedKeyMaterialImpl)

        state = await self.__get_state()

        # The key exchange can be a network instance here, but it can't be a migrated instance, so we don't
        # have to worry about that here.
        if key_exchange.is_network_instance():
            # Perform lookup of the signed pre key and pre key public keys in case the header is not filled
            signed_pre_keys_by_id = { v: k for k, v in (await self.__get_signed_pre_key_ids()).items() }
            if key_exchange.signed_pre_key_id not in signed_pre_keys_by_id:
                raise KeyExchangeFailed(f"No signed pre key with id {key_exchange.signed_pre_key_id} known.")

            pre_keys_by_id = { v: k for k, v in (await self.__get_pre_key_ids()).items() }
            if key_exchange.pre_key_id not in pre_keys_by_id:
                raise KeyExchangeFailed(f"No pre key with id {key_exchange.pre_key_id} known.")

            # Update the key exchange information with the filled header
            key_exchange = KeyExchangeImpl(
                x3dh.Header(
                    key_exchange.header.identity_key,
                    key_exchange.header.ephemeral_key,
                    signed_pre_keys_by_id[key_exchange.signed_pre_key_id],
                    pre_keys_by_id[key_exchange.pre_key_id]
                ),
                key_exchange.signed_pre_key_id,
                key_exchange.pre_key_id
            )

        try:
            shared_secret, associated_data, signed_pre_key = await state.get_shared_secret_passive(
                key_exchange.header
            )
        except x3dh.KeyAgreementException as e:
            raise KeyExchangeFailed() from e

        try:
            double_ratchet, decrypted_message = await DoubleRatchetImpl.decrypt_initial_message(
                diffie_hellman_ratchet_curve25519.DiffieHellmanRatchet,
                RootChainKDFImpl,
                MessageChainKDFImpl,
                DoubleRatchetImpl.MESSAGE_CHAIN_CONSTANT,
                self.max_num_per_message_skipped_keys,
                self.max_num_per_session_skipped_keys,
                AEADImpl,
                shared_secret,
                signed_pre_key.priv,
                encrypted_key_material.encrypted_message,
                associated_data
            )
        except Exception as e:
            raise DecryptionFailed(
                "Decryption of the initial message as part of passive session building failed."
            ) from e

        session = SessionImpl(
            bare_jid,
            device_id,
            Initiation.PASSIVE,
            key_exchange,
            associated_data,
            double_ratchet
        )

        plain_key_material = PlainKeyMaterialImpl(
            decrypted_message[:PlainKeyMaterialImpl.KEY_LENGTH],
            decrypted_message[PlainKeyMaterialImpl.KEY_LENGTH:]
        )

        return session, plain_key_material

    async def encrypt_plaintext(self, plaintext: bytes) -> Tuple[ContentImpl, PlainKeyMaterialImpl]:
        # Generate KEY_LENGTH bytes of cryptographically secure random data for the key
        key = secrets.token_bytes(PlainKeyMaterialImpl.KEY_LENGTH)

        # Generate 12 bytes for the IV
        initialization_vector = secrets.token_bytes(12)

        # Encrypt the plaintext using AES-128 (the 128 bit are implied by the key size) in GCM mode and the
        # previously created key and IV
        aes = Cipher(
            algorithms.AES(key),
            modes.GCM(initialization_vector),
            backend=default_backend()
        ).encryptor()
        ciphertext = aes.update(plaintext) + aes.finalize()

        # This authentication tag is not truncated
        auth_tag = aes.tag

        return ContentImpl(ciphertext, initialization_vector), PlainKeyMaterialImpl(key, auth_tag)

    async def encrypt_empty(self) -> Tuple[ContentImpl, PlainKeyMaterialImpl]:
        return ContentImpl.make_empty(), PlainKeyMaterialImpl.make_empty()

    async def encrypt_key_material(
        self,
        session: Session,
        plain_key_material: PlainKeyMaterial
    ) -> EncryptedKeyMaterialImpl:
        assert isinstance(session, SessionImpl)
        assert isinstance(plain_key_material, PlainKeyMaterialImpl)

        # Rebuild the associated data such that it follows the order
        # sender identity key || recipient identity key
        # regardless of who initiated the session. This is to conform to the undocumented behaviour of
        # libsignal.
        associated_data = session.associated_data
        if session.initiation is Initiation.PASSIVE:
            associated_data = (
                associated_data[StateImpl.IDENTITY_KEY_ENCODING_LENGTH:]
                + associated_data[:StateImpl.IDENTITY_KEY_ENCODING_LENGTH]
            )

        return EncryptedKeyMaterialImpl(
            session.bare_jid,
            session.device_id,
            await session.double_ratchet.encrypt_message(
                plain_key_material.key + plain_key_material.auth_tag,
                associated_data
            )
        )

    async def decrypt_plaintext(self, content: Content, plain_key_material: PlainKeyMaterial) -> bytes:
        assert isinstance(content, ContentImpl)
        assert isinstance(plain_key_material, PlainKeyMaterialImpl)

        assert not content.empty

        # Decrypt the plaintext using AES-128 (the 128 bit are implied by the key size) in GCM mode and the
        # key and IV in plain_key_material, while also verifying the authentication tag.
        aes = Cipher(
            algorithms.AES(plain_key_material.key),
            modes.GCM(content.initialization_vector, plain_key_material.auth_tag),
            backend=default_backend()
        ).decryptor()
        try:
            return aes.update(content.ciphertext) + aes.finalize()
        except InvalidTag as e:
            raise DecryptionFailed("Ciphertext decryption failed.") from e

    async def decrypt_key_material(
        self,
        session: Session,
        encrypted_key_material: EncryptedKeyMaterial
    ) -> PlainKeyMaterialImpl:
        assert isinstance(session, SessionImpl)
        assert isinstance(encrypted_key_material, EncryptedKeyMaterialImpl)

        # Rebuild the associated data such that it follows the order
        # sender identity key || recipient identity key
        # regardless of who initiated the session. This is to conform to the undocumented behaviour of
        # libsignal.
        associated_data = session.associated_data
        if session.initiation is Initiation.ACTIVE:
            associated_data = (
                associated_data[StateImpl.IDENTITY_KEY_ENCODING_LENGTH:]
                + associated_data[:StateImpl.IDENTITY_KEY_ENCODING_LENGTH]
            )

        try:
            decrypted_message = await session.double_ratchet.decrypt_message(
                encrypted_key_material.encrypted_message,
                associated_data
            )
        except Exception as e:
            raise DecryptionFailed("Key material decryption failed.") from e

        session.confirm()

        return PlainKeyMaterialImpl(
            decrypted_message[:PlainKeyMaterialImpl.KEY_LENGTH],
            decrypted_message[PlainKeyMaterialImpl.KEY_LENGTH:]
        )

    async def signed_pre_key_age(self) -> int:
        return (await self.__get_state()).signed_pre_key_age()

    async def rotate_signed_pre_key(self) -> None:
        state = await self.__get_state()

        state.rotate_signed_pre_key()

        await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

    async def hide_pre_key(self, session: Session) -> bool:
        assert isinstance(session, SessionImpl)

        # This method is only called with KeyExchangeImpl instances that have the pre key byte data set. We do
        # not have to worry about the field containing a filler value and the assertion is merely there to
        # satisfy the type system.
        assert session.key_exchange.header.pre_key is not None

        state = await self.__get_state()

        hidden = state.hide_pre_key(session.key_exchange.header.pre_key)

        await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

        return hidden

    async def delete_pre_key(self, session: Session) -> bool:
        assert isinstance(session, SessionImpl)

        # This method is only called with KeyExchangeImpl instances that have the pre key byte data set. We do
        # not have to worry about the field containing a filler value and the assertion is merely there to
        # satisfy the type system.
        assert session.key_exchange.header.pre_key is not None

        state = await self.__get_state()

        deleted = state.delete_pre_key(session.key_exchange.header.pre_key)

        await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

        return deleted

    async def delete_hidden_pre_keys(self) -> None:
        state = await self.__get_state()

        state.delete_hidden_pre_keys()

        await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

    async def get_num_visible_pre_keys(self) -> int:
        return (await self.__get_state()).get_num_visible_pre_keys()

    async def generate_pre_keys(self, num_pre_keys: int) -> None:
        state = await self.__get_state()

        state.generate_pre_keys(num_pre_keys)

        await self.__storage.store(f"/{self.namespace}/x3dh", state.json)

    async def get_bundle(self, bare_jid: str, device_id: int) -> BundleImpl:
        bundle = (await self.__get_state()).bundle

        return BundleImpl(
            bare_jid,
            device_id,
            bundle,
            (await self.__get_signed_pre_key_ids())[bundle.signed_pre_key],
            {
                pre_key: pre_key_id
                for pre_key, pre_key_id
                in (await self.__get_pre_key_ids()).items()
                if pre_key in bundle.pre_keys
            }
        )

    @property
    def supports_labels(self) -> bool:
        return False

    async def sign_own_label(self, label: str) -> bytes:
        raise NotImplementedError("This OMEMO backend does not support device labels.")

    async def verify_label_signature(self, label: str, signature: bytes, identity_key: bytes) -> bool:
        raise NotImplementedError("This OMEMO backend does not support device labels.")

    async def purge(self) -> None:
        for bare_jid in (await self.__storage.load_list(f"/{self.namespace}/bare_jids", str)).maybe([]):
            await self.purge_bare_jid(bare_jid)

        await self.__storage.delete(f"/{self.namespace}/bare_jids")
        await self.__storage.delete(f"/{self.namespace}/x3dh")
        await self.__storage.delete(f"/{self.namespace}/signed_pre_key_ids")
        await self.__storage.delete(f"/{self.namespace}/pre_key_ids")
        await self.__storage.delete(f"/{self.namespace}/pre_key_id_counter")

    async def purge_bare_jid(self, bare_jid: str) -> None:
        storage = self.__storage

        for device_id in (await storage.load_list(f"/{self.namespace}/{bare_jid}/device_ids", int)).maybe([]):
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/initiation")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/identity_key")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/ephemeral_key")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/signed_pre_key")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/signed_pre_key_id")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/pre_key")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/key_exchange/pre_key_id")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/associated_data")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/double_ratchet")
            await storage.delete(f"/{self.namespace}/{bare_jid}/{device_id}/confirmed")

        await storage.delete(f"/{self.namespace}/{bare_jid}/device_ids")

        bare_jids = set((await storage.load_list(f"/{self.namespace}/bare_jids", str)).maybe([]))
        bare_jids.remove(bare_jid)
        await storage.store(f"/{self.namespace}/bare_jids", list(bare_jids))

    async def __get_signed_pre_key_ids(self) -> Dict[bytes, int]:
        """
        Assigns an id to each signed pre key currently available in the X3DH state, both the current signed
        pre key and the old signed pre key that is kept around for one more rotation period. Once assigned to
        a signed pre key, its id will never change.

        Returns:
            The mapping from signed pre key to id.
        """

        state = await self.__get_state()

        signed_pre_key = state.bundle.signed_pre_key
        old_signed_pre_key = state.old_signed_pre_key

        # Load the existing signed pre key ids from the storage
        signed_pre_key_ids = {
            base64.b64decode(signed_pre_key_b64): signed_pre_key_id
            for signed_pre_key_b64, signed_pre_key_id
            in (await self.__storage.load_dict(
                f"/{self.namespace}/signed_pre_key_ids",
                int
            )).maybe({}).items()
        }

        # Take note of the highest id that was assigned, default to 0 if no ids were assigned yet
        signed_pre_key_id_counter = max(
            signed_pre_key_id
            for _, signed_pre_key_id
            in signed_pre_key_ids.items()
        ) if len(signed_pre_key_ids) > 0 else 0

        # Prepare the dictionary to hold updated signed pre key ids
        new_signed_pre_key_ids: Dict[bytes, int] = {}

        # Assign the next highest id to the signed pre key, if there is no id assigned to it yet.
        signed_pre_key_id_counter += 1
        new_signed_pre_key_ids[signed_pre_key] = signed_pre_key_ids.get(
            signed_pre_key,
            signed_pre_key_id_counter
        )

        # Assign the next highest id to the old signed pre key, if there is no id assigned to it yet. This
        # should never happen, since the old signed pre key should have been assigned an id when it was the
        # (non-old) signed pre key, however there might be edge cases of the signed pre key rotating twice
        # before the assigned ids are updated.
        if old_signed_pre_key is not None:
            signed_pre_key_id_counter += 1
            new_signed_pre_key_ids[old_signed_pre_key] = signed_pre_key_ids.get(
                old_signed_pre_key,
                signed_pre_key_id_counter
            )

        # If the ids have changed, store them
        if new_signed_pre_key_ids != signed_pre_key_ids:
            await self.__storage.store(f"/{self.namespace}/signed_pre_key_ids", {
                base64.b64encode(signed_pre_key).decode("ASCII"): signed_pre_key_id
                for signed_pre_key, signed_pre_key_id
                in new_signed_pre_key_ids.items()
            })

        return new_signed_pre_key_ids

    async def __get_pre_key_ids(self) -> Dict[bytes, int]:
        """
        Assigns an id to each pre key currently available in the X3DH state, both hidden and visible pre keys.
        Once assigned to a pre key, its id will never change.

        Returns:
            The mapping from pre key to id.
        """

        state = await self.__get_state()

        pre_keys = state.bundle.pre_keys | state.hidden_pre_keys

        # Load the existing pre key ids from the storage
        pre_key_ids = {
            base64.b64decode(pre_key_b64): pre_key_id
            for pre_key_b64, pre_key_id
            in (await self.__storage.load_dict(f"/{self.namespace}/pre_key_ids", int)).maybe({}).items()
        }

        # Load the pre key id counter from the storage
        pre_key_id_counter = (await self.__storage.load_primitive(
            f"/{self.namespace}/pre_key_id_counter",
            int
        )).maybe(0)

        # Prepare the dictionary to hold updated pre key ids
        new_pre_key_ids: Dict[bytes, int] = {}

        # Assign the next highest id to each pre key if there is no existing id assigned to it
        for pre_key in pre_keys:
            pre_key_id_counter += 1
            new_pre_key_ids[pre_key] = pre_key_ids.get(pre_key, pre_key_id_counter)

        # If the ids have changed, store them
        if new_pre_key_ids != pre_key_ids:
            await self.__storage.store(f"/{self.namespace}/pre_key_ids", {
                base64.b64encode(pre_key).decode("ASCII"): pre_key_id
                for pre_key, pre_key_id
                in new_pre_key_ids.items()
            })

            await self.__storage.store(f"/{self.namespace}/pre_key_id_counter", pre_key_id_counter)

        return new_pre_key_ids



import base64
from typing import Optional, Set, Tuple, cast
import xml.etree.ElementTree as ET

from omemo import (
    DeviceList,
    DeviceListDownloadFailed,
    EncryptedKeyMaterial,
    KeyExchange,
    Message,
    SenderNotFound,
    SessionManager
)
import x3dh
import xeddsa
try:
    import xmlschema
except ImportError as import_error:
    raise ImportError(
        "Optional dependency xmlschema not found. Please install xmlschema, or install this package using"
        " `pip install oldmemo[xml]`, to use the ElementTree-based XML serialization/parser helpers."
    ) from import_error

from .oldmemo import NAMESPACE, BundleImpl, ContentImpl, EncryptedKeyMaterialImpl, KeyExchangeImpl, StateImpl


__all__ = [
    "serialize_device_list",
    "parse_device_list",
    "serialize_bundle",
    "parse_bundle",
    "serialize_message",
    "parse_message"
]


NS = f"{{{NAMESPACE}}}"


DEVICE_LIST_SCHEMA = xmlschema.XMLSchema("""<?xml version="1.0" encoding="utf8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
    targetNamespace="eu.siacs.conversations.axolotl"
    xmlns="eu.siacs.conversations.axolotl">

    <xs:element name="device">
        <xs:complexType>
            <xs:attribute name="id" type="xs:unsignedInt" use="required"/>
        </xs:complexType>
    </xs:element>

    <xs:element name="list">
        <xs:complexType>
            <xs:sequence minOccurs="0" maxOccurs="unbounded">
                <xs:element ref="device"/>
            </xs:sequence>
        </xs:complexType>
    </xs:element>
</xs:schema>
""")


BUNDLE_SCHEMA = xmlschema.XMLSchema("""<?xml version="1.0" encoding="utf8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
    targetNamespace="eu.siacs.conversations.axolotl"
    xmlns="eu.siacs.conversations.axolotl">

    <xs:element name="signedPreKeyPublic">
        <xs:complexType>
            <xs:simpleContent>
                <xs:extension base="xs:base64Binary">
                    <xs:attribute name="signedPreKeyId" type="xs:unsignedInt"/>
                </xs:extension>
            </xs:simpleContent>
        </xs:complexType>
    </xs:element>

    <xs:element name="signedPreKeySignature" type="xs:base64Binary"/>
    <xs:element name="identityKey" type="xs:base64Binary"/>

    <xs:element name="preKeyPublic">
        <xs:complexType>
            <xs:simpleContent>
                <xs:extension base="xs:base64Binary">
                    <xs:attribute name="preKeyId" type="xs:unsignedInt" use="required"/>
                </xs:extension>
            </xs:simpleContent>
        </xs:complexType>
    </xs:element>

    <xs:element name="prekeys">
        <xs:complexType>
            <xs:sequence>
                <xs:element ref="preKeyPublic" maxOccurs="unbounded"/>
            </xs:sequence>
        </xs:complexType>
    </xs:element>

    <xs:element name="bundle">
        <xs:complexType>
            <xs:all>
                <xs:element ref="signedPreKeyPublic"/>
                <xs:element ref="signedPreKeySignature"/>
                <xs:element ref="identityKey"/>
                <xs:element ref="prekeys"/>
            </xs:all>
        </xs:complexType>
    </xs:element>
</xs:schema>
""")


MESSAGE_SCHEMA = xmlschema.XMLSchema11("""<?xml version="1.1" encoding="utf8"?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
    targetNamespace="eu.siacs.conversations.axolotl"
    xmlns="eu.siacs.conversations.axolotl">

    <xs:element name="payload" type="xs:base64Binary"/>
    <xs:element name="iv" type="xs:base64Binary"/>

    <xs:element name="key">
        <xs:complexType>
            <xs:simpleContent>
                <xs:extension base="xs:base64Binary">
                    <xs:attribute name="rid" type="xs:unsignedInt" use="required"/>
                    <xs:attribute name="prekey" type="xs:boolean"/>
                </xs:extension>
            </xs:simpleContent>
        </xs:complexType>
    </xs:element>

    <xs:element name="header">
        <xs:complexType>
            <xs:all>
                <xs:element ref="key" minOccurs="0" maxOccurs="unbounded"/>
                <xs:element ref="iv"/>
            </xs:all>
            <xs:attribute name="sid" type="xs:unsignedInt"/>
        </xs:complexType>
    </xs:element>

    <xs:element name="encrypted">
        <xs:complexType>
            <xs:all>
                <xs:element ref="header"/>
                <xs:element ref="payload" minOccurs="0"/>
            </xs:all>
        </xs:complexType>
    </xs:element>
</xs:schema>
""")


def serialize_device_list(device_list: DeviceList) -> ET.Element:
    """
    Args:
        device_list: The device list to serialize. The first entry of each tuple is the device id, and the
            second entry is the optional signed label. Note that labels are not supported by this version of
            the specification and will not be included in the XML.

    Returns:
        The serialized device list as an XML element.
    """

    list_elt = ET.Element(f"{NS}list")

    for device_id, _ in device_list.items():
        device_elt = ET.SubElement(list_elt, f"{NS}device")
        device_elt.set("id", str(device_id))

    return list_elt


def parse_device_list(element: ET.Element) -> DeviceList:
    """
    Args:
        element: The XML element to parse the device list from.

    Returns:
        The extracted device list. The first entry of each tuple is the device id, and the second entry is the
        optional signed label. Note that labels are not supported by this version of the specification, so all
        labels will be set to ``None``.

    Raises:
        xmlschema.XMLSchemaValidationError: in case the element does not conform to the XML schema given in
            the specification.
    """

    DEVICE_LIST_SCHEMA.validate(element)

    return { int(cast(str, device_elt.get("id"))): None for device_elt in element.iter(f"{NS}device") }


def serialize_bundle(bundle: BundleImpl) -> ET.Element:
    """
    Args:
        bundle: The bundle to serialize.

    Returns:
        The serialized bundle as an XML element.
    """

    bundle_elt = ET.Element(f"{NS}bundle")

    signed_pre_key_signature_mut = bytearray(bundle.bundle.signed_pre_key_sig)
    signed_pre_key_signature_mut[63] |= bundle.bundle.identity_key[31] & 0x80
    signed_pre_key_signature = bytes(signed_pre_key_signature_mut)

    ET.SubElement(
        bundle_elt,
        f"{NS}signedPreKeyPublic",
        attrib={ "signedPreKeyId": str(bundle.signed_pre_key_id) }
    ).text = base64.b64encode(StateImpl.serialize_public_key(bundle.bundle.signed_pre_key)).decode("ASCII")

    ET.SubElement(
        bundle_elt,
        f"{NS}signedPreKeySignature"
    ).text = base64.b64encode(signed_pre_key_signature).decode("ASCII")

    ET.SubElement(
        bundle_elt,
        f"{NS}identityKey"
    ).text = base64.b64encode(StateImpl.serialize_public_key(xeddsa.ed25519_pub_to_curve25519_pub(
        bundle.bundle.identity_key
    ))).decode("ASCII")

    prekeys_elt = ET.SubElement(bundle_elt, f"{NS}prekeys")
    for pre_key in bundle.bundle.pre_keys:
        ET.SubElement(
            prekeys_elt,
            f"{NS}preKeyPublic",
            attrib={ "preKeyId": str(bundle.pre_key_ids[pre_key]) }
        ).text = base64.b64encode(StateImpl.serialize_public_key(pre_key)).decode("ASCII")

    return bundle_elt


def parse_bundle(element: ET.Element, bare_jid: str, device_id: int) -> BundleImpl:
    """
    Args:
        element: The XML element to parse the bundle from.
        bare_jid: The bare JID this bundle belongs to.
        device_id: The device id of the specific device this bundle belongs to.

    Returns:
        The extracted bundle.

    Raises:
        ValueError: in case of malformed data that still passed the schema validation.
        xmlschema.XMLSchemaValidationError: in case the element does not conform to the XML schema given in
            the specification.
    """

    BUNDLE_SCHEMA.validate(element)

    spkp_elt = cast(ET.Element, element.find(f"{NS}signedPreKeyPublic"))
    pkp_elts = list(element.iter(f"{NS}preKeyPublic"))

    signed_pre_key_signature = base64.b64decode(cast(str, cast(ET.Element, element.find(
        f"{NS}signedPreKeySignature"
    )).text))

    identity_key = xeddsa.curve25519_pub_to_ed25519_pub(StateImpl.parse_public_key(base64.b64decode(
        cast(str, cast(ET.Element, element.find(f"{NS}identityKey")).text)
    )), bool((signed_pre_key_signature[63] >> 7) & 1))

    signed_pre_key_signature_mut = bytearray(signed_pre_key_signature)
    signed_pre_key_signature_mut[63] &= 0x7f
    signed_pre_key_signature = bytes(signed_pre_key_signature_mut)

    pre_key_ids = {
        StateImpl.parse_public_key(base64.b64decode(cast(str, pkp_elt.text))):
            int(cast(str, pkp_elt.get("preKeyId")))
        for pkp_elt
        in pkp_elts
    }

    return BundleImpl(
        bare_jid,
        device_id,
        x3dh.Bundle(
            identity_key,
            StateImpl.parse_public_key(base64.b64decode(cast(str, spkp_elt.text))),
            signed_pre_key_signature,
            frozenset(pre_key_ids.keys())
        ),
        int(cast(str, spkp_elt.get("signedPreKeyId"))),
        pre_key_ids
    )


def serialize_message(message: Message) -> ET.Element:
    """
    Args:
        message: The message to serialize.

    Returns:
        The serialized message as an XML element.
    """

    assert isinstance(message.content, ContentImpl)

    encrypted_elt = ET.Element(f"{NS}encrypted")

    header_elt = ET.SubElement(encrypted_elt, f"{NS}header", attrib={ "sid": str(message.device_id) })

    for encrypted_key_material, key_exchange in message.keys:
        assert isinstance(encrypted_key_material, EncryptedKeyMaterialImpl)

        key_elt = ET.SubElement(
            header_elt,
            f"{NS}key",
            attrib={ "rid": str(encrypted_key_material.device_id) }
        )

        authenticated_message = encrypted_key_material.serialize()

        if key_exchange is None:
            key_elt.text = base64.b64encode(authenticated_message).decode("ASCII")
        else:
            assert isinstance(key_exchange, KeyExchangeImpl)

            key_exchange_serialized, _sign_bit_set = key_exchange.serialize(authenticated_message)

            key_elt.set("prekey", "true")
            key_elt.text = base64.b64encode(key_exchange_serialized).decode("ASCII")

    ET.SubElement(
        header_elt,
        f"{NS}iv"
    ).text = base64.b64encode(message.content.initialization_vector).decode("ASCII")

    if not message.content.empty:
        ET.SubElement(
            encrypted_elt,
            f"{NS}payload"
        ).text = base64.b64encode(message.content.ciphertext).decode("ASCII")

    return encrypted_elt


async def parse_message(
    element: ET.Element,
    sender_bare_jid: str,
    own_bare_jid: str,
    session_manager: SessionManager
) -> Message:
    """
    Args:
        element: The XML element to parse the message from.
        sender_bare_jid: The bare JID of the sender.
        own_bare_jid: The bare JID of the XMPP account decrypting this message, i.e. us.
        session_manager: The session manager instance is required to find one piece of information that the
            oldmemo message serialization format lacks with regards to the identity key.

    Returns:
        The extracted message.

    Raises:
        ValueError: in case there is malformed data not caught be the XML schema validation.
        xmlschema.XMLSchemaValidationError: in case the element does not conform to the XML schema given in
            the specification.
        SenderNotFound: in case the public information about the sending device could not be found or is
            incomplete.

    Warning:
        This version of the OMEMO specification matches key material to recipient purely by device id. The
        library, however, matches by bare JID and device id. Since the XML doesn't include the bare JID, the
        structures expected by the library can't be filled correctly. Instead, to match the behaviour of the
        specification, the bare JID of all key material included in the message is assigned to
        ``own_bare_jid``, i.e. our own bare JID, which achieves the desired effect of matching only on
        the device id.
    """

    MESSAGE_SCHEMA.validate(element)

    payload_elt = element.find(f"{NS}payload")
    header_elt = cast(ET.Element, element.find(f"{NS}header"))
    iv_elt = header_elt.find(f"{NS}iv")
    sender_device_id = int(cast(str, header_elt.get("sid")))

    # The following code might seem overkill just to find the sign of the identity key. However, to make usage
    # of the library as simple as possible, I believe it is worth it. Since most results are cached, this
    # shouldn't significantly slow down the overall decryption flow either.
    sender_device = next((
        device
        for device
        in await session_manager.get_device_information(sender_bare_jid)
        if device.device_id == sender_device_id
    ), None)

    if sender_device is None:
        try:
            # If the device wasn't found, refresh the device list
            await session_manager.refresh_device_list(NAMESPACE, sender_bare_jid)
        except DeviceListDownloadFailed as e:
            raise SenderNotFound(
                "Couldn't find public information about the device which sent this message and an attempt to"
                " refresh the sender's device list failed."
            ) from e

        sender_device = next((
            device
            for device
            in await session_manager.get_device_information(sender_bare_jid)
            if device.device_id == sender_device_id
        ), None)

    if sender_device is None:
        raise SenderNotFound(
            "Couldn't find public information about the device which sent this message. I.e. the device"
            " either does not appear in the device list of the sending XMPP account, or the bundle of the"
            " sending device could not be downloaded."
        )

    set_sign_bit = bool((sender_device.identity_key[31] >> 7) & 1)

    keys: Set[Tuple[EncryptedKeyMaterial, Optional[KeyExchange]]] = set()

    for key_elt in header_elt.iter(f"{NS}key"):
        recipient_device_id = int(cast(str, key_elt.get("rid")))
        content = base64.b64decode(cast(str, key_elt.text))

        key_exchange: Optional[KeyExchangeImpl] = None
        authenticated_message: bytes
        if key_elt.get("prekey", "false") in [ "true", "1" ]:
            key_exchange, authenticated_message = KeyExchangeImpl.parse(content, set_sign_bit)
        else:
            authenticated_message = content

        encrypted_key_material = EncryptedKeyMaterialImpl.parse(
            authenticated_message,
            own_bare_jid,
            recipient_device_id
        )

        keys.add((encrypted_key_material, key_exchange))

    return Message(
        NAMESPACE,
        sender_bare_jid,
        sender_device_id,
        (
            ContentImpl.make_empty()
            if payload_elt is None or iv_elt is None
            else ContentImpl(
                base64.b64decode(cast(str, payload_elt.text)),
                base64.b64decode(cast(str, iv_elt.text))
            )
        ),
        frozenset(keys)
    )



# This import from future (theoretically) enables sphinx_autodoc_typehints to handle type aliases better
from __future__ import annotations

from abc import ABC, abstractmethod
import base64
from typing import Awaitable, Callable, Dict, List, Optional, Set
from typing_extensions import TypedDict

import doubleratchet
import omemo
from omemo.session import Initiation
import x3dh
import x3dh.migrations
import x3dh.types
import xeddsa

from .oldmemo import NAMESPACE, BundleImpl, StateImpl


__all__ = [
    "OwnData",
    "Trust",
    "Session",
    "BoundOTPK",
    "StateSuper",
    "State",
    "LegacyStorage",
    "migrate"
]


class OwnData(TypedDict):
    """
    This TypedDict describes how the own data was expected to be returned by the corresponding legacy storage
    method.
    """

    own_bare_jid: str
    own_device_id: int


class Trust(TypedDict):
    """
    This TypedDict describes how trust information was expected to be returned by the corresponding legacy
    storage method.
    """

    key: str
    trusted: bool


class Session(TypedDict):
    """
    This TypedDict describes how session instances (more precisely ``ExtendedDoubleRatchet`` instances) were
    serialized in the pre-stable serialization format.
    """

    super: doubleratchet.JSONObject
    other_ik: str


class BoundOTPK(TypedDict):
    """
    Used as part of the legacy state format to represent a bound pre key.
    """

    otpk: str
    id: int


class StateSuper(TypedDict):
    """
    Used as part of the legacy state format to represent the super class of the ``X3DHDoubleRatchet``.
    """

    super: x3dh.JSONObject
    spk_id: int
    spk_pub: Optional[str]
    otpk_id_counter: int
    otpk_ids: Dict[str, int]


class State(TypedDict):
    """
    This TypedDict describes how the state (more precisely ``X3DHDoubleRatchet``) was serialized in the
    pre-stable serialization format. Note that the ``pk_messages`` entry has been omitted from this type since
    it is not needed for migration. The same applies to the ``version`` field, which apparently never had any
    relevance.
    """

    super: StateSuper
    bound_otpks: Dict[str, Dict[int, BoundOTPK]]


class LegacyStorage(ABC):
    # pylint: disable=invalid-name
    """
    This is a slightly modified copy of the storage interface used by legacy (i.e. pre-1.0.0) python-omemo.
    All methods related to storing values have been removed. A few methods for deleting values have been added
    instead. Methods related to efficient bulk loading have been removed as well.
    """

    @abstractmethod
    async def loadOwnData(self) -> Optional[OwnData]:
        """
        Returns:
            The own data stored in this instance, if any.
        """

    @abstractmethod
    async def deleteOwnData(self) -> None:
        """
        Delete the own data stored in this instance, if any. Do not raise if there is none.
        """

    @abstractmethod
    async def loadState(self) -> Optional[State]:
        """
        Returns:
            The state stored in this instance, if any.
        """

    @abstractmethod
    async def deleteState(self) -> None:
        """
        Delete the state stored in this instance, if any. Do not raise if there is none.
        """

    @abstractmethod
    async def loadSession(self, bare_jid: str, device_id: int) -> Optional[Session]:
        """
        Args:
            bare_jid: The bare JID.
            device_id: The device id.

        Returns:
            The session stored in this instance for the given bare JID + device id, if any.
        """

    @abstractmethod
    async def deleteSession(self, bare_jid: str, device_id: int) -> None:
        """
        Delete the session stored in this instance for the given bare JID + device id, if any. Do not raise if
        there is none.

        Args:
            bare_jid: The bare JID.
            device_id: The device id.
        """

    @abstractmethod
    async def loadActiveDevices(self, bare_jid: str) -> Optional[List[int]]:
        """
        Args:
            bare_jid: The bare JID.

        Returns:
            The list of active devices stored in this instance for the given bare JID, if any.

        Note:
            It doesn't matter whether you return ``None`` or an empty list of no list is stored for this bare
            JID.
        """

    @abstractmethod
    async def loadInactiveDevices(self, bare_jid: str) -> Optional[Dict[int, int]]:
        """
        Args:
            bare_jid: The bare JID.

        Returns:
            A mapping of inactive devices stored in this instance for the given bare JID, if any. The mapping
            maps from device id to the timestamp of last activity (seconds since epoch).

        Note:
            It doesn't matter whether you return ``None`` or an empty dictionary of no dictionary is stored
            for this bare JID.
        """

    @abstractmethod
    async def deleteActiveDevices(self, bare_jid: str) -> None:
        """
        Delete the list of active devices stored in this instance for the given bare JID, if any. Do not raise
        if there is none.

        Args:
            bare_jid: The bare JID.
        """

    @abstractmethod
    async def deleteInactiveDevices(self, bare_jid: str) -> None:
        """
        Delete the dictionary of inactive devices stored in this instance for the given bare JID, if any. Do
        not raise if there is none.

        Args:
            bare_jid: The bare JID.
        """

    @abstractmethod
    async def loadTrust(self, bare_jid: str, device_id: int) -> Optional[Trust]:
        """
        Args:
            bare_jid: The bare JID.
            device_id: The device id.

        Returns:
            The trust information stored in this instance for the given bare JID + device id, if any.
        """

    @abstractmethod
    async def deleteTrust(self, bare_jid: str, device_id: int) -> None:
        """
        Delete the trust information stored in this instance for the given bare JID + device id, if any. Do
        not raise if there is none.

        Args:
            bare_jid: The bare JID.
            device_id: The device id.
        """

    @abstractmethod
    async def listJIDs(self) -> Optional[List[str]]:
        """
        Returns:
            A list of all bare JIDs that have associated device lists stored in the storage, if any. For a
            bare JID to be included in the list, it doesn't matter if the associated device lists are empty or
            not. Return ``None`` if the list of bare JIDs has been deleted, do not return an empty list in
            that case.
        """

    @abstractmethod
    async def deleteJIDList(self) -> None:
        """
        Delete the list of bare JIDs as returned by :meth:`listJIDs`, if it exists. Do not raise if it
        doesn't.
        """


async def migrate(
    legacy_storage: LegacyStorage,
    storage: omemo.Storage,
    trusted_trust_level_name: str,
    undecided_trust_level_name: str,
    untrusted_trust_level_name: str,
    download_bundle: Callable[[str, int], Awaitable[BundleImpl]]
) -> None:
    """
    Migrate the data from a legacy storage instance to the current storage format. This function is
    idempotent, which means that you can run it without checking whether migrations are required or not. If
    migrations are not required, the function will do nothing. This function also makes sure that only safely
    migrated data is deleted from the legacy storage, such that migration failure at any point leaves both
    storages in a consistent state.

    Args:
        legacy_storage: The legacy storage instance to migrate the data from. This assumes that the storage
            was used with legacy (pre-1.0.0) python-omemo in conjunction with python-omemo-backend-signal. If
            the storage was used with a backend other than python-omemo-backend-signal, automated migration is
            not possible with this function.
        storage: The storage implementation to migrate the data to.
        trusted_trust_level_name: The legacy storage format stored trust as a boolean, i.e. there were only
            trusted or untrusted devices. This is the name of the custom trust level to store when a trusted
            device is migrated.
        undecided_trust_level_name: The name of the custom trust level to store when a device without any
            associated trust information is migrated.
        untrusted_trust_level_name: The name of the custom trust level to store when an untrusted device is
            migrated.
        download_bundle: A function which downloads the bundle of the given bare JID + device id. May raise
            :class:`~omemo.session_manager.BundleDownloadFailed` (or a subclass thereof) to indicate failure.
    """

    # python-omemo SessionManager
    # - f"/devices/{bare_jid}/{device_id}/namespaces" List[str]
    # - f"/devices/{bare_jid}/{device_id}/active" Dict[str, bool]
    # - f"/devices/{bare_jid}/{device_id}/label" Optional[str]
    # - f"/devices/{bare_jid}/{device_id}/identity_key" bytes
    # - f"/devices/{bare_jid}/list" List[int]
    # - f"/trust/{bare_jid}/{base64.urlsafe_b64encode(identity_key).decode('ASCII')}" str
    # - "/own_device_id" int

    # python-omemo IdentityKeyPair
    # - "/ikp/is_seed" bool
    # - "/ikp/key" 32 bytes

    # python-oldmemo
    # - f"/{self.namespace}/x3dh" x3dh.JSONObject
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/initiation" str
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/identity_key" bytes
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/ephemeral_key" bytes
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/signed_pre_key" bytes
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/signed_pre_key_id" int
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/pre_key" bytes
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/key_exchange/pre_key_id" int
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/associated_data" bytes
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/double_ratchet" doubleratchet.JSONObject
    # - f"/{self.namespace}/{session.bare_jid}/{session.device_id}/confirmed" bool
    # - f"/{self.namespace}/bare_jids" List[str]
    # - f"/{self.namespace}/{session.bare_jid}/device_ids" List[int]
    # - f"/{self.namespace}/signed_pre_key_ids" Dict[str, int]
    # - f"/{self.namespace}/pre_key_ids" Dict[str, int]
    # - f"/{self.namespace}/pre_key_id_counter" int

    # The own data is the very first thing to be loaded and the very last thing to be deleted.
    own_data = await legacy_storage.loadOwnData()
    if own_data is None:
        return

    # The own bare JID isn't part of the new storage format.
    await storage.store("/own_device_id", own_data["own_device_id"])

    # The legacy state format contains both the X3DH state and information required for the migration of
    # sessions
    state = await legacy_storage.loadState()
    if state is None:
        return

    # Migrate the X3DH state
    await storage.store(f"/{NAMESPACE}/x3dh", state["super"]["super"])

    # Now, load and migrate the own identity key since it's needed for session migrations later on. The own
    # identity key is part of the X3DH state. Parse the X3DH state and extract the identity key.
    base_state_model, _ = x3dh.migrations.parse_base_state_model(state["super"]["super"])

    # The base state model contains the secret and the secret type of the identity key
    own_identity_key_secret = base_state_model.identity_key.secret
    own_identity_key_secret_type = base_state_model.identity_key.secret_type

    # Migrate the secret
    await storage.store_bytes("/ikp/key", own_identity_key_secret)

    # The type of secret should be x3dh.types.SecretType.PRIV, since that's the only format supported by
    # legacy python-omemo + python-omemo-backend-signal, but it doesn't hurt to check the type anyway.
    await storage.store("/ikp/is_seed", own_identity_key_secret_type is x3dh.types.SecretType.SEED)

    # From the secret, we can calculate the Ed25519 public key
    own_identity_key = (
        xeddsa.seed_to_ed25519_pub(own_identity_key_secret)
        if own_identity_key_secret_type is x3dh.types.SecretType.SEED
        else xeddsa.priv_to_ed25519_pub(own_identity_key_secret)
    )

    # Prepare the identity key in the serialized format required to build the associated data later as part of
    # session migration
    own_identity_key_serialized = StateImpl.serialize_public_key(xeddsa.ed25519_pub_to_curve25519_pub(
        own_identity_key
    ))

    # With the state loaded and identity key prepared, begin with the migration of information related to
    # other devices, i.e. device information including trust and sessions.
    bare_jids = await legacy_storage.listJIDs()
    if bare_jids is not None:
        for bare_jid in bare_jids:
            # Load active and inactive devices of this bare JID
            active_devices = await legacy_storage.loadActiveDevices(bare_jid)
            inactive_devices = await legacy_storage.loadInactiveDevices(bare_jid)

            # The timestamp on the inactive devices is not needed
            active_device_ids = set() if active_devices is None else set(active_devices)
            inactive_device_ids = set() if inactive_devices is None else set(inactive_devices.keys())

            # Migrate general device information
            migrated_devices: Set[int] = set()
            for device_id in active_device_ids | inactive_device_ids:
                active = device_id in active_device_ids

                # At this point, there are two options: either, trust information for the device exists,
                # including the identity key of the device in Curve25519 format, or it doesn't. Either way,
                # there is a problem. The new format expects the identity key in Ed25519 format. To convert a
                # Curve25519 key to Ed25519, the missing sign bit has to be provided. The sign bit can be
                # fetched from the signed pre key signature included in the public bundle of the device. That
                # means, if the trust information including the Curve25519 identity is present, the bundle has
                # to be fetched for the sign bit. If the trust information is not present, the bundle has to
                # be fetched for the whole identity key. That means, either way, we need the bundle here.
                # The single exception is the own device, since out own identity key is available.
                identity_key: bytes

                if bare_jid == own_data["own_bare_jid"] and device_id == own_data["own_device_id"]:
                    identity_key = own_identity_key
                else:
                    try:
                        bundle = await download_bundle(bare_jid, device_id)
                    except omemo.BundleDownloadFailed:
                        # A device whose bundle can't be fetched cannot be migrated. Delete all data related
                        # to the device and skip to the next one.
                        await legacy_storage.deleteSession(bare_jid, device_id)
                        await legacy_storage.deleteTrust(bare_jid, device_id)
                        continue

                    # The BundleImpl structure contains the identity in Ed25519 form, thus no further
                    # conversion is required. The bundle parsing code has already taken care of extracting the
                    # sign bit from the signed pre key signature and converting the key from Curve25519
                    # format.
                    identity_key = bundle.identity_key

                # Load the trust information stored for this device.
                trust = await legacy_storage.loadTrust(bare_jid, device_id)

                # If trust information is available, verify that the identity key stored for the device
                # matches the one just fetched from the bundle
                if trust is not None:
                    legacy_storage_identity_key = base64.b64decode(trust["key"])

                    if xeddsa.ed25519_pub_to_curve25519_pub(identity_key) != legacy_storage_identity_key:
                        # If the stored identity key doesn't match the bundle identity key, the device is not
                        # migrated. Delete all data related to the device and skip to the next one.
                        await legacy_storage.deleteSession(bare_jid, device_id)
                        await legacy_storage.deleteTrust(bare_jid, device_id)
                        continue

                # Select the custom trust level name to assign to the device upon migration
                trust_level_name = undecided_trust_level_name
                if trust is not None:
                    if trust["trusted"]:
                        trust_level_name = trusted_trust_level_name
                    else:
                        trust_level_name = untrusted_trust_level_name

                # All available data about this device has been gathered, migrate it
                await storage.store(f"/devices/{bare_jid}/{device_id}/namespaces", [ NAMESPACE ])
                await storage.store(f"/devices/{bare_jid}/{device_id}/active", { NAMESPACE: active })
                await storage.store(f"/devices/{bare_jid}/{device_id}/label", None)
                await storage.store_bytes(f"/devices/{bare_jid}/{device_id}/identity_key", identity_key)
                await storage.store(
                    f"/trust/{bare_jid}/{base64.urlsafe_b64encode(identity_key).decode('ASCII')}",
                    trust_level_name
                )

                # The device has been migrated successfully, add it to the corresponding set but don't delete
                # its data yet.
                migrated_devices.add(device_id)

            # Write the list of successfully migrated devices
            await storage.store(f"/devices/{bare_jid}/list", list(migrated_devices))

            # Look for sessions with the successfully migrated devices and migrate those too
            migrated_sessions: Set[int] = set()
            for device_id in migrated_devices:
                session = await legacy_storage.loadSession(bare_jid, device_id)
                if session is None:
                    continue

                legacy_storage_identity_key = base64.b64decode(session["other_ik"])
                double_ratchet = session["super"]

                # Same situation as above: the identity key is only included in Curve25519 format, but is
                # needed in Ed25519 format to become part of the key exchange information. Luckily, we have
                # already requested the bundle above and stored the identity key in Ed25519 format. Load it
                # here (from cache) and make sure it matches the identity key stored in legacy storage.
                identity_key = (await storage.load_bytes(
                    f"/devices/{bare_jid}/{device_id}/identity_key"
                )).from_just()

                if xeddsa.ed25519_pub_to_curve25519_pub(identity_key) != legacy_storage_identity_key:
                    # If the stored identity key doesn't match the bundle identity key, the session is not
                    # migrated. Delete all data related to it and skip to the next one.
                    await legacy_storage.deleteSession(bare_jid, device_id)
                    continue

                # If the identity keys match, store the Ed25519 one.
                await storage.store_bytes(
                    f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/identity_key",
                    identity_key
                )

                # Prepare the serialized identity as needed to build the associated data byte string
                identity_key_serialized = StateImpl.serialize_public_key(
                    xeddsa.ed25519_pub_to_curve25519_pub(identity_key)
                )

                # A bunch of information stored in the new storage format wasn't directly available in the
                # legacy format, namely which party initiated the session, whether the session was confirmed
                # via the passive party responding, and the key exchange information that was used to build
                # the session. All of this information is used for protocol stability reasons regarding the
                # initial X3DH key agreement. Some of this information can be extracted/guessed from
                # "bound pre keys" that were used by the legacy format and stored with the state. Bound pre
                # keys are pre keys associated with passively initiated sessions upon creation, and deleted at
                # a user-defined point in the future, hopefully as soon as the session has been fully
                # confirmed and no more key exchange messages are to be expected. All implementations I know
                # of didn't delete bound pre keys at all, which is good for the migration. The presence of a
                # bound pre key indicates that a session was passively initiated and allows us to restore at
                # least part of the key exchange information. Sessions without an associated bound pre key are
                # migrated as actively initiated and confirmed, which is the constellation of parameters that
                # makes sure the missing key exchange information is never accessed.
                bound_otpk = state["bound_otpks"].get(bare_jid, {}).get(device_id)
                if bound_otpk is None:
                    # Set the initiation to active as explained above.
                    await storage.store(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/initiation",
                        Initiation.ACTIVE.name
                    )

                    # Store the associated data in the format expected for actively initated sessions.
                    await storage.store_bytes(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/associated_data",
                        own_identity_key_serialized + identity_key_serialized
                    )

                    # Without a bound pre key, the pre key and pre key id fields of the key exchange can't be
                    # set correctly. The KeyExchangeImpl class detects and handles the following filler values
                    # though:
                    await storage.store(f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/pre_key_id", -1)
                    await storage.store_bytes(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/pre_key",
                        b"\x00" * 32
                    )
                else:
                    # Set the initiation to passive as explained above.
                    await storage.store(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/initiation",
                        Initiation.PASSIVE.name
                    )

                    # Store the associated data in the format expected for passively initated sessions.
                    await storage.store_bytes(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/associated_data",
                        identity_key_serialized + own_identity_key_serialized
                    )

                    # With a bound pre key, we can at least fill the pre key and pre key id fields of the key
                    # exchange information correctly.
                    await storage.store(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/pre_key_id",
                        bound_otpk["id"]
                    )
                    await storage.store_bytes(
                        f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/pre_key",
                        base64.b64decode(bound_otpk["otpk"])
                    )

                # All sessions are marked as confirmed. This makes sure that the code never attempts to send
                # the incomplete key exchange information.
                await storage.store(f"/{NAMESPACE}/{bare_jid}/{device_id}/confirmed", True)

                # The serialized double ratchet just has to be assigned to the correct key.
                await storage.store(f"/{NAMESPACE}/{bare_jid}/{device_id}/double_ratchet", double_ratchet)

                # The ephemeral key, signed pre key and signed pre key ids used during key exchange are
                # unknown. The values in storage still have to be filled though. The KeyExchangeImpl class
                # detects and handles the following filler values correctly:
                await storage.store_bytes(
                    f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/ephemeral_key",
                    b"\x00" * 32
                )

                await storage.store(f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/signed_pre_key_id", -1)
                await storage.store_bytes(
                    f"/{NAMESPACE}/{bare_jid}/{device_id}/key_exchange/signed_pre_key",
                    b"\x00" * 32
                )

                # The session was migrated successfully, add it to the corresponding set but don't delete its
                # data yet.
                migrated_sessions.add(device_id)

            # Write the list of successfully migrated sessions
            await storage.store(f"/{NAMESPACE}/{bare_jid}/device_ids", list(migrated_sessions))

            # Migration completed for this bare JID, delete all legacy data to avoid double migration
            await legacy_storage.deleteActiveDevices(bare_jid)
            await legacy_storage.deleteInactiveDevices(bare_jid)
            for device_id in migrated_devices:
                await legacy_storage.deleteTrust(bare_jid, device_id)
            for device_id in migrated_sessions:
                await legacy_storage.deleteSession(bare_jid, device_id)

        # All bare JIDs have been migrated, write the list of bare JIDs...
        await storage.store(f"/{NAMESPACE}/bare_jids", list(bare_jids))

        # ...and delete the list of JIDs
        await legacy_storage.deleteJIDList()

    # What remains to be migrated are (signed) pre key id mappings and counters.

    # The legacy format didn't keep the old signed pre key around, so there is at most the current signed pre
    # key id to migrate, which is optional in the legacy storage format too.
    if state["super"]["spk_pub"] is not None:
        await storage.store(
            f"/{NAMESPACE}/signed_pre_key_ids",
            { state["super"]["spk_pub"]: state["super"]["spk_id"] }
        )

    # The pre key id mapping and id counter are already in the format/types required by the new storage
    # format, thus it's simply a matter of migrating them to their new keys.
    await storage.store(f"/{NAMESPACE}/pre_key_ids", state["super"]["otpk_ids"])
    await storage.store(f"/{NAMESPACE}/pre_key_id_counter", state["super"]["otpk_id_counter"])

    # Finally delete the legacy state and the own data.
    await legacy_storage.deleteState()
    await legacy_storage.deleteOwnData()
